@@ -1,10 +1,11 @@
+import json
 import logging
 import random
-import requests
 import time
 
 from patroni.dcs.zookeeper import ZooKeeper
-from requests.exceptions import RequestException
+from patroni.request import get as requests_get
+from patroni.utils import uri
 
 logger = logging.getLogger(__name__)
 
@@ -18,7 +19,7 @@ class ExhibitorEnsembleProvider(object):
         self._uri_path = uri_path
         self._poll_interval = poll_interval
         self._exhibitors = hosts
-        self._master_exhibitors = hosts
+        self._boot_exhibitors = hosts
         self._zookeeper_hosts = ''
         self._next_poll = None
         while not self.poll():
@@ -31,7 +32,7 @@ class ExhibitorEnsembleProvider(object):
 
         json = self._query_exhibitors(self._exhibitors)
         if not json:
-            json = self._query_exhibitors(self._master_exhibitors)
+            json = self._query_exhibitors(self._boot_exhibitors)
 
         if isinstance(json, dict) and 'servers' in json and 'port' in json:
             self._next_poll = time.time() + self._poll_interval
@@ -46,12 +47,11 @@ class ExhibitorEnsembleProvider(object):
     def _query_exhibitors(self, exhibitors):
         random.shuffle(exhibitors)
         for host in exhibitors:
-            uri = 'http://{0}:{1}{2}'.format(host, self._exhibitor_port, self._uri_path)
             try:
-                response = requests.get(uri, timeout=self.TIMEOUT)
-                return response.json()
-            except RequestException:
-                pass
+                response = requests_get(uri('http', (host, self._exhibitor_port), self._uri_path), timeout=self.TIMEOUT)
+                return json.loads(response.data.decode('utf-8'))
+            except Exception:
+                logging.debug('Request to %s failed', host)
         return None
 
     @property
@@ -68,7 +68,7 @@ class Exhibitor(ZooKeeper):
         config['hosts'] = self._ensemble_provider.zookeeper_hosts
         super(Exhibitor, self).__init__(config)
 
-    def _load_cluster(self):
+    def _load_cluster(self, path, loader):
         if self._ensemble_provider.poll():
             self._client.set_hosts(self._ensemble_provider.zookeeper_hosts)
-        return super(Exhibitor, self)._load_cluster()
+        return super(Exhibitor, self)._load_cluster(path, loader)
